/********************************************************************************/
/* lu_bench.c:                                                                  */
/* Copyright (C) 2014 Tomonori Kouya                                            */
/*                                                                              */
/* This program is free software: you can redistribute it and/or modify it      */
/* under the terms of the GNU Lesser General Public License as published by the */
/* Free Software Foundation, either version 3 of the License or any later       */
/* version.                                                                     */
/*                                                                              */
/* This program is distributed in the hope that it will be useful, but WITHOUT  */
/* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or        */
/* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License */
/* for more details.                                                            */
/*                                                                              */
/* You should have received a copy of the GNU Lesser General Public License     */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>.        */
/*                                                                              */
/********************************************************************************/
/* stdio.h */
#ifndef _STDIO_H
#include <stdio.h>
#endif

/* math.h */
#ifndef _MATH_H
#include <math.h>
#endif

#include "bnc.h"

#ifdef USE_IMKL
	#include "mkl.h"
	#include "mkl_cblas.h" // for Intel Math Kernel Library
#endif

#include "matmul_strassen.h"


// (1) L11 * U11 = A11
//
//    start_index
// +--+-----+--------+
// |  +-----+        |
// |  |\ U11|        |
// |  |L11\ |        |
// +  +-----+        |
// |  min_dim        |
// |                 |
// |                 |
// |                 |
// +-----------+-----+
//
int DLUdecomp_square(DMatrix a, long int start_index, long int min_dim)
{
	long int i, j, k, imax, jmax, itmp, dim;
	double dtmp, dmaxii;

	dim = a->col_dim;

	imax = start_index + min_dim;
	if(imax > dim) imax = dim;

	jmax = imax;

	// dim > min_dim
	for(i = start_index; i < imax; i++)
	{
		dmaxii = fabs(get_dmatrix_ij(a, i, i));
		if(dmaxii == 0.0)
		{
			fprintf(stderr, "%ld : Error! (DLUdecomp_square)!\n", i);
			return -1;
		}

		for(j = (i + 1); j < jmax; j++)
			set_dmatrix_ij(a, j, i, get_dmatrix_ij(a, j, i) / get_dmatrix_ij(a, i, i));

		for(j = (i + 1); j < jmax; j++)
		{
			for(k = (i + 1); k < jmax; k++)
				set_dmatrix_ij(a, j, k, get_dmatrix_ij(a, j, k) - get_dmatrix_ij(a, j, i) * get_dmatrix_ij(a, i, k));
		}
	}

	return 0;
}

// (2) Solve L21 * U11 = A21
//
//    start_index
// +--+-----+--------+
// |  +-----+        |
// |  |\ U11|        |
// |  |L11\ |        |
// +  +-----+        |
// |  |     |        |
// |  |     |        |
// |  | L21 |        |
// |  |     |        |
// +--+-----+--------+
//
int DLUdecomp_l21(DMatrix l21, DMatrix a, long int start_index, long int min_dim)
{
	long int i, j, k, imax, jmax, itmp, dim;
	double dtmp, dmaxii;

	dim = a->col_dim;

	imax = start_index + min_dim;
	if(imax > dim) imax = dim;

	jmax = imax;

	// dim > min_dim
	for(i = imax; i < dim; i++)
	{
		dmaxii = fabs(get_dmatrix_ij(a, i, i));
		if(dmaxii == 0.0)
		{
			fprintf(stderr, "%ld : Error! (DLUdecomp_l21)!\n", i);
			return -1;
		}

		for(j = start_index; j < jmax; j++)
		{
			dtmp = get_dmatrix_ij(a, i, j);
			for(k = start_index; k < j; k++)
			{
				dtmp -= get_dmatrix_ij(a, i, k) * get_dmatrix_ij(a, k, j);
			}
			dtmp /= get_dmatrix_ij(a, j, j);
			
			//printf("(i - start_index, j - imax) = (%ld, %ld), %ld, %ld\n", i - imax, j - start_index, l21->row_dim, l21->col_dim);
			//printf("(i              , j       ) = (%ld, %ld) %25.17e, %25.17e\n", i, j, dtmp, get_dmatrix_ij(a, j, j));
			set_dmatrix_ij(a  , i              , j       , dtmp);
			set_dmatrix_ij(l21, i - imax, j - start_index, dtmp);
		}
	}

	return 0;
}

// (3) Solve L11 * U12 = A21
//
//    start_index
// +--+-----+--------+
// |  +-----+--------+
// |  |\ U11|  U12   |
// |  |L11\ |        |
// +  +-----+--------+
// |  |     |        |
// |  |     |        |
// |  | L21 |        |
// |  |     |        |
// +--+-----+--------+
//
int DLUdecomp_u12(DMatrix u12, DMatrix a, long int start_index, long int min_dim)
{
	long int i, j, k, imax, jmax, itmp, dim;
	double dtmp, dmaxii;

	dim = a->col_dim;

	imax = start_index + min_dim;
	if(imax > dim) imax = dim;

	jmax = imax;

	// dim > min_dim
	for(i = imax; i < dim; i++)
	{
		dmaxii = fabs(get_dmatrix_ij(a, i, i));
		if(dmaxii == 0.0)
		{
			fprintf(stderr, "%ld : Error! (DLUdecomp_u12)!\n", i);
			return -1;
		}

		for(j = start_index; j < jmax; j++)
		{
			dtmp = get_dmatrix_ij(a, j, i);
			//printf("(j              , i              ) = (%ld, %ld) %25.17e\n", j, i , dtmp);
			for(k = start_index; k < j; k++)
			{
				dtmp -= get_dmatrix_ij(a, j, k) * get_dmatrix_ij(a, k, i);
				//printf("(j - start_index, k - start_index) = (%ld, %ld)\n", j - start_index, k - start_index);
			}
			//printf("(j              , i              ) = (%ld, %ld) %25.17e\n", j, i , dtmp);
			set_dmatrix_ij(a  , j              , i       , dtmp);
			set_dmatrix_ij(u12, j - start_index, i - imax, dtmp);
		}
	}

	return 0;
}

// (4) D22 := L21 * U12
// (5) A22 := A22 - D22
//
//    start_index
// +--+-----+--------+
// |  +-----+--------+
// |  |\ U11|  U12   |
// |  |L11\ |        |
// +  +-----+--------+
// |  |     |        |
// |  |     |        |
// |  | L21 |  A22   |
// |  |     |        |
// +--+-----+--------+
//
int DLUdecomp_a22(DMatrix a, DMatrix d22, DMatrix l21, DMatrix u12, long int start_index, long int min_dim)
{
	long int i, j, k, imax, jmax, itmp, dim, index[4], d22_index[4];
	double dtmp, dmaxii;

	dim = a->col_dim;

	imax = start_index + min_dim;
	if(imax > dim) imax = dim;

	jmax = imax;

	// d22 := l21 * u12
#ifdef USE_IMKL
	cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans, (lapack_int)(l21->row_dim), (lapack_int)(u12->col_dim), (lapack_int)(l21->col_dim), 1.0, l21->element, (lapack_int)(l21->col_dim), u12->element, (lapack_int)(u12->col_dim), 0.0, d22->element, (d22->col_dim));
#else
	mul_dmatrix_simple(d22, l21, u12);
#endif

	// a22 := a22 - d22
	index[0] = imax;
	index[1] = dim;
	index[2] = jmax;
	index[3] = dim;
	d22_index[0] = 0;
	d22_index[1] = d22->row_dim;
	d22_index[2] = 0;
	d22_index[3] = d22->col_dim;

	sub_dmatrix_partial(a, index, a, index, d22, d22_index);

	return 0;
}


/************************************************************/
/*                                                          */
/*           LU Decomposition for Square Dense Matrix       */
/*                                 (Double Precision)       */
/*                                                          */
/*                 ver. 0.0 1997.10.24 (Fri) Tomonori Kouya */
/*                                                          */
/************************************************************/
int DLUdecomp_strassen(DMatrix a, long int min_dim)
/************************************************************/
/*                                                          */
/* entries                                                  */
/*       DMatrix a: Matrix (given by user)                 */
/*                   a[i * dim + j] ... (i,j)-entry of A    */
/*                                                          */
/* returns                                                  */
/*       a[]: LU decomposed matrix                          */
/*                                                          */
/************************************************************/
{
	long int i, row_dim, col_dim, dim;
	double dtmp, dmaxii;
	DMatrix l21, u12, d22;

	dim = a->col_dim;

	// dim <= min_dim
	if(dim <= min_dim)
	{
		DLUdecomp(a);
		return 0;
	}

	// dim > min_dim
	for(i = 0; i < dim; i += min_dim)
	{
		// Initialize
		row_dim = dim - (i + min_dim);
		if(row_dim <= 0) 
		{
			// (1) L11 * U11 = A11
			if((dim - i) > 0)
				DLUdecomp_square(a, i, dim - i);

			break;
		}

		col_dim = row_dim;

		//printf("%2ld : (row_dim, col_dim, min_dim) = (%ld, %ld, %ld)\n", i, row_dim, col_dim, min_dim);

		// (1) L11 * U11 = A11
		DLUdecomp_square(a, i, min_dim);

		//printf("i = %ld\n", i);
		//print_dmatrix(a);

		l21 = init_dmatrix(row_dim, min_dim);
		u12 = init_dmatrix(min_dim, col_dim);
		d22 = init_dmatrix(row_dim, col_dim);

		// (2) Solve L21 * U11 = A21
		DLUdecomp_l21(l21, a, i, min_dim);
		//print_dmatrix(a);
		//printf("L21:\n");
		//print_dmatrix(l21);

		// (3) Solve L11 * U12 = A12
		DLUdecomp_u12(u12, a, i, min_dim);
		//print_dmatrix(a);
		//printf("U12:\n");
		//print_dmatrix(u12);

		//printf("A:\n");
		//print_dmatrix(a);

		// (4) D22 := L21 * U12
		// (5) A22 := A22 - D22
		DLUdecomp_a22(a, d22, l21, u12, i, min_dim);
		//printf("D22:\n");
		//print_dmatrix(d22);
		//print_dmatrix(a);

		// clear
		free_dmatrix(l21);
		free_dmatrix(u12);
		free_dmatrix(d22);

	}

	return 0;
}

// n!
double dfactorial(long int n)
{
	double ret;

	ret = 1.0;

	if(n > 0)
	{
		do
		{
			ret *= (double)n;
		} while(n-- > 1);
	}

	return ret;
}

// Pascal Matrix
void pascal_dmatrix(DMatrix ret, long int dim)
{
	long int i, j;
	double element;

	for(i = 0; i < ret->row_dim; i++)
	{
		for(j = 0; j < ret->col_dim; j++)
		{
			element = dfactorial(i + j) / (dfactorial(i) * dfactorial(j));
			set_dmatrix_ij(ret, i, j, element);
		}
	}
}

void get_dproblem(DMatrix a, DVector b, DVector ans)
{
	long int i, j, k;
	double tmp;

	/* Lotkin Matrix */
/*	for(i = 0; i < a->col_dim; i++)
		set_dmatrix_ij(a, 0, i, 1.0);
	for(i = 1; i < a->row_dim; i++)
	{
		for(j = 0; j < a->col_dim; j++)
			set_dmatrix_ij(a, i, j, 1.0 / (i + j + 1));
	}
*/
	frank_dmatrix(a, a->row_dim);

	/* Answer */
	for(i = 0; i < ans->dim; i++)
		set_dvector_i(ans, i, (double)i);

	/* Make constant vector */
	mul_dmatrix_dvec(b, a, ans);
}

#ifdef USE_GMP

// (1) L11 * U11 = A11
//
//    start_index
// +--+-----+--------+
// |  +-----+        |
// |  |\ U11|        |
// |  |L11\ |        |
// +  +-----+        |
// |  min_dim        |
// |                 |
// |                 |
// |                 |
// +-----------+-----+
//
int MPFLUdecomp_square(MPFMatrix a, long int start_index, long int min_dim)
{
	long int i, j, k, imax, jmax, itmp, dim;
	mpf_t dtmp, dmaxii;

	dim = a->col_dim;

	imax = start_index + min_dim;
	if(imax > dim) imax = dim;

	jmax = imax;

	mpf_init2(dtmp, a->prec);
	mpf_init2(dmaxii, a->prec);

	// dim > min_dim
	for(i = start_index; i < imax; i++)
	{
		mpf_abs(dmaxii, get_mpfmatrix_ij(a, i, i));
		if(mpf_cmp_ui(dmaxii, 0UL) == 0)
		{
			fprintf(stderr, "%ld : Error! (MPFLUdecomp_square)!\n", i);
			mpf_clear(dtmp);
			mpf_clear(dmaxii);
			return -1;
		}

		for(j = (i + 1); j < jmax; j++)
		{
			mpf_div(dtmp, get_mpfmatrix_ij(a, j, i), get_mpfmatrix_ij(a, i, i));
			set_mpfmatrix_ij(a, j, i, dtmp);
		}

		for(j = (i + 1); j < jmax; j++)
		{
			for(k = (i + 1); k < jmax; k++)
			{
				mpf_mul(dtmp, get_mpfmatrix_ij(a, j, i), get_mpfmatrix_ij(a, i, k));
				mpf_sub(dtmp, get_mpfmatrix_ij(a, j, k), dtmp);
				set_mpfmatrix_ij(a, j, k, dtmp);
			}
		}
	}

	mpf_clear(dtmp);
	mpf_clear(dmaxii);

	return 0;
}

// (2) Solve L21 * U11 = A21
//
//    start_index
// +--+-----+--------+
// |  +-----+        |
// |  |\ U11|        |
// |  |L11\ |        |
// +  +-----+        |
// |  |     |        |
// |  |     |        |
// |  | L21 |        |
// |  |     |        |
// +--+-----+--------+
//
int MPFLUdecomp_l21(MPFMatrix l21, MPFMatrix a, long int start_index, long int min_dim)
{
	long int i, j, k, imax, jmax, itmp, dim;
	mpf_t dtmp, dmaxii;

	dim = a->col_dim;

	imax = start_index + min_dim;
	if(imax > dim) imax = dim;

	jmax = imax;

	mpf_init2(dmaxii, l21->prec);
	mpf_init2(dtmp, l21->prec);

	// dim > min_dim
	for(i = imax; i < dim; i++)
	{
		mpf_abs(dmaxii, get_mpfmatrix_ij(a, i, i));
		if(mpf_cmp_ui(dmaxii, 0UL) == 0)
		{
			fprintf(stderr, "%ld : Error! (MPFLUdecomp_l21)!\n", i);
			mpf_clear(dtmp);
			mpf_clear(dmaxii);
			return -1;
		}

		for(j = start_index; j < jmax; j++)
		{
			mpf_set(dtmp, get_mpfmatrix_ij(a, i, j));
			for(k = start_index; k < j; k++)
			{
				//dtmp -= get_dmatrix_ij(a, i, k) * get_dmatrix_ij(a, k, j);
				mpf_mul(dmaxii, get_mpfmatrix_ij(a, i, k), get_mpfmatrix_ij(a, k, j));
				mpf_sub(dtmp, dtmp, dmaxii);
			}
			mpf_div(dtmp, dtmp, get_mpfmatrix_ij(a, j, j));
			
			//printf("(i - start_index, j - imax) = (%ld, %ld), %ld, %ld\n", i - imax, j - start_index, l21->row_dim, l21->col_dim);
			//printf("(i              , j       ) = (%ld, %ld) %25.17e, %25.17e\n", i, j, dtmp, get_dmatrix_ij(a, j, j));
			set_mpfmatrix_ij(a  , i              , j       , dtmp);
			set_mpfmatrix_ij(l21, i - imax, j - start_index, dtmp);
		}
	}

	mpf_clear(dtmp);
	mpf_clear(dmaxii);

	return 0;
}

// (3) Solve L11 * U12 = A21
//
//    start_index
// +--+-----+--------+
// |  +-----+--------+
// |  |\ U11|  U12   |
// |  |L11\ |        |
// +  +-----+--------+
// |  |     |        |
// |  |     |        |
// |  | L21 |        |
// |  |     |        |
// +--+-----+--------+
//
int MPFLUdecomp_u12(MPFMatrix u12, MPFMatrix a, long int start_index, long int min_dim)
{
	long int i, j, k, imax, jmax, itmp, dim;
	mpf_t dtmp, dmaxii;

	dim = a->col_dim;

	imax = start_index + min_dim;
	if(imax > dim) imax = dim;

	jmax = imax;

	mpf_init2(dmaxii, u12->prec);
	mpf_init2(dtmp, u12->prec);

	// dim > min_dim
	for(i = imax; i < dim; i++)
	{
		mpf_abs(dmaxii, get_mpfmatrix_ij(a, i, i));
		if(mpf_cmp_ui(dmaxii, 0UL) == 0)
		{
			fprintf(stderr, "%ld : Error! (MPFLUdecomp_u12)!\n", i);
			mpf_clear(dtmp);
			mpf_clear(dmaxii);
			return -1;
		}

		for(j = start_index; j < jmax; j++)
		{
			mpf_set(dtmp, get_mpfmatrix_ij(a, j, i));
			//printf("(j              , i              ) = (%ld, %ld) %25.17e\n", j, i , dtmp);
			for(k = start_index; k < j; k++)
			{
				//dtmp -= get_dmatrix_ij(a, j, k) * get_dmatrix_ij(a, k, i);
				mpf_mul(dmaxii, get_mpfmatrix_ij(a, j, k), get_mpfmatrix_ij(a, k, i));
				mpf_sub(dtmp, dtmp, dmaxii);
				//printf("(j - start_index, k - start_index) = (%ld, %ld)\n", j - start_index, k - start_index);
			}
			//printf("(j              , i              ) = (%ld, %ld) %25.17e\n", j, i , dtmp);
			set_mpfmatrix_ij(a  , j              , i       , dtmp);
			set_mpfmatrix_ij(u12, j - start_index, i - imax, dtmp);
		}
	}

	mpf_clear(dtmp);
	mpf_clear(dmaxii);

	return 0;
}

// (4) D22 := L21 * U12
// (5) A22 := A22 - D22
//
//    start_index
// +--+-----+--------+
// |  +-----+--------+
// |  |\ U11|  U12   |
// |  |L11\ |        |
// +  +-----+--------+
// |  |     |        |
// |  |     |        |
// |  | L21 |  A22   |
// |  |     |        |
// +--+-----+--------+
//
#define STRASSEN_MIN_DIM 32
int MPFLUdecomp_a22(MPFMatrix a, MPFMatrix d22, MPFMatrix l21, MPFMatrix u12, long int start_index, long int min_dim)
{
	long int i, j, k, imax, jmax, itmp, dim, index[4], d22_index[4];
	MPFVector diag_left, diag_right;

	dim = a->col_dim;

	imax = start_index + min_dim;
	if(imax > dim) imax = dim;

	jmax = imax;

	// d22 := l21 * u12
#if defined(USE_STRASSEN) || defined(USE_WINOGRAD)
#ifdef USE_SCALING
	diag_left = init2_mpfvector(l21->row_dim, d22->prec);
	diag_right = init2_mpfvector(u12->col_dim, d22->prec);

	left_scaling_mpfmatrix(l21, l21, diag_left, NULL);
	right_scaling_mpfmatrix(u12, u12, diag_right, NULL);
#endif

//	mul_mpfmatrix_strassen(d22, l21, u12, min_dim);
//	mul_mpfmatrix_strassen(d22, l21, u12, min_dim / 2);
//	mul_mpfmatrix_strassen(d22, l21, u12, min_dim / 4);
	mul_mpfmatrix_strassen(d22, l21, u12, STRASSEN_MIN_DIM);

#ifdef USE_SCALING
	mul_mpfmatrix_mpfdiag(d22, diag_left, 0, d22, diag_right, 0); 

	mul_mpfmatrix_mpfdiag(l21, diag_left, 1, l21, NULL, 0);
	mul_mpfmatrix_mpfdiag(u12, NULL, 0, u12, diag_right, 1);

	free_mpfvector(diag_left);
	free_mpfvector(diag_right);
#endif

#elif USE_BLOCK
	mul_mpfmatrix_block(d22, l21, u12, STRASSEN_MIN_DIM);
#else
	mul_mpfmatrix_simple(d22, l21, u12);
#endif

	// a22 := a22 - d22
	index[0] = imax;
	index[1] = dim;
	index[2] = jmax;
	index[3] = dim;
	d22_index[0] = 0;
	d22_index[1] = d22->row_dim;
	d22_index[2] = 0;
	d22_index[3] = d22->col_dim;

	sub_mpfmatrix_partial(a, index, a, index, d22, d22_index);

	return 0;
}


/************************************************************/
/*                                                          */
/*           LU Decomposition for Square Dense Matrix       */
/*                                 (Double Precision)       */
/*                                                          */
/*                 ver. 0.0 1997.10.24 (Fri) Tomonori Kouya */
/*                                                          */
/************************************************************/
int MPFLUdecomp_strassen(MPFMatrix a, long int min_dim)
/************************************************************/
/*                                                          */
/* entries                                                  */
/*       MPFMatrix a: Matrix (given by user)                 */
/*                   a[i * dim + j] ... (i,j)-entry of A    */
/*                                                          */
/* returns                                                  */
/*       a[]: LU decomposed matrix                          */
/*                                                          */
/************************************************************/
{
	long int i, row_dim, col_dim, dim;
	MPFMatrix l21, u12, d22;

	dim = a->col_dim;

	// dim <= min_dim
	if(dim <= min_dim)
	{
		MPFLUdecomp(a);
		return 0;
	}

	// dim > min_dim
	for(i = 0; i < dim; i += min_dim)
	{
		// Initialize
		row_dim = dim - (i + min_dim);
		if(row_dim <= 0) 
		{
			// (1) L11 * U11 = A11
			if((dim - i) > 0)
				MPFLUdecomp_square(a, i, dim - i);

			break;
		}

		col_dim = row_dim;

		//printf("%2ld : (row_dim, col_dim, min_dim) = (%ld, %ld, %ld)\n", i, row_dim, col_dim, min_dim);

		// (1) L11 * U11 = A11
		MPFLUdecomp_square(a, i, min_dim);

		//printf("i = %ld\n", i);
		//print_dmatrix(a);

/*		l21 = init2_mpfmatrix(row_dim, min_dim, a->prec + (a->prec / 4));
		u12 = init2_mpfmatrix(min_dim, col_dim, a->prec + (a->prec / 4));
		d22 = init2_mpfmatrix(row_dim, col_dim, a->prec + (a->prec / 4));
*/		l21 = init2_mpfmatrix(row_dim, min_dim, a->prec);
		u12 = init2_mpfmatrix(min_dim, col_dim, a->prec);
		d22 = init2_mpfmatrix(row_dim, col_dim, a->prec);

		// (2) Solve L21 * U11 = A21
		MPFLUdecomp_l21(l21, a, i, min_dim);
		//print_dmatrix(a);
		//printf("L21:\n");
		//print_dmatrix(l21);

		// (3) Solve L11 * U12 = A12
		MPFLUdecomp_u12(u12, a, i, min_dim);
		//print_dmatrix(a);
		//printf("U12:\n");
		//print_dmatrix(u12);

		//printf("A:\n");
		//print_dmatrix(a);

		// (4) D22 := L21 * U12
		// (5) A22 := A22 - D22
		MPFLUdecomp_a22(a, d22, l21, u12, i, min_dim);
		//printf("D22:\n");
		//print_dmatrix(d22);
		//print_dmatrix(a);

		// clear
		free_mpfmatrix(l21);
		free_mpfmatrix(u12);
		free_mpfmatrix(d22);

	}

	return 0;
}


// n!
void mpf_factorial(mpf_t ret, long int n)
{
	mpf_set_ui(ret, 1UL);

	if(n > 0)
	{
		do
		{
			mpf_mul_ui(ret, ret, n);;
		} while(n-- > 1);
	}

	return;
}

// Pascal Matrix
void pascal_mpfmatrix(MPFMatrix ret, long int dim)
{
	long int i, j;
	mpf_t element, tmp;

	mpf_init2(element, ret->prec);
	mpf_init2(tmp, ret->prec);

	for(i = 0; i < ret->row_dim; i++)
	{
		for(j = 0; j < ret->col_dim; j++)
		{
			//element = dfactorial(i + j) / (dfactorial(i) * dfactorial(j));
			mpf_factorial(element, i + j);
			mpf_factorial(tmp, i);
			mpf_mul(element, element, tmp);
			mpf_factorial(tmp, j);
			mpf_mul(element, element, tmp);

			set_mpfmatrix_ij(ret, i, j, element);
		}
	}

	mpf_clear(element);
	mpf_clear(tmp);
}

// I - randmatrix
void im_rand_mpfmatrix(MPFMatrix ret, unsigned long seed)
{
	long int i, j;
	mpf_t element;

	mpf_init2(element, ret->prec);

	// set seed
	srand(seed);

	// element := (-1)^rand() * rand() / RAND_MAX;
	for(i = 0; i < ret->row_dim; i++)
	{
		for(j = 0; j < ret->col_dim; j++)
		{
			//printf("%ld == %ld\n", i, j);
			mpf_set_d(element, (double)(rand()));
			mpf_div_ui(element, element, (unsigned long)RAND_MAX);
			if((rand() % 2) != 0)
				mpf_neg(element, element);

			set_mpfmatrix_ij(ret, i, j, element);
		}

		//printf("%ld == %ld\n", i, i);
		// element[i][i] := 1 - element[i][i];
		mpf_ui_sub(element, 1UL, get_mpfmatrix_ij(ret, i, i));
		set_mpfmatrix_ij(ret, i, i, element);
	}

	mpf_clear(element);
}

void get_mpfproblem(MPFMatrix a, MPFVector b, MPFVector ans)
{
	long int i, j, k;
	mpf_t tmp;

	mpf_init(tmp);

	/* Lotkin Matrix */
/*	for(i = 0; i < a->col_dim; i++)
		set_mpfmatrix_ij_d(a, 0, i, 1.0);
	for(i = 1; i < a->row_dim; i++)
	{
		for(j = 0; j < a->col_dim; j++)
		{
			mpf_set_ui(tmp, 1UL);
			mpf_div_ui(tmp, tmp, (unsigned long)(i + j + 1));
			set_mpfmatrix_ij(a, i, j, tmp);
		}
	}
*/

#ifdef USE_LOTKIN
	lotkin_mpfmatrix(a, a->row_dim);
#elif USE_IM_RAND
	im_rand_mpfmatrix(a, 10UL);
#else
	frank_mpfmatrix(a, a->row_dim);
//	pascal_mpfmatrix(a, a->row_dim);
#endif
	/* Answer */
	for(i = 0; i < ans->dim; i++)
	{
		mpf_set_si(tmp, i);
		set_mpfvector_i(ans, i, tmp);
	}

	/* Make constant vector */
	mul_mpfmatrix_mpfvec(b, a, ans);
}
#endif


//#define DIM 128
//#define DIM 128
//#define DIM 1024

int main(int argc, char *argv[])
{
	long int dim;
	double stime, etime[6];
	DMatrix da;
	DVector db, dx, dans;
#ifdef USE_GMP
	unsigned long prec;
	long int block_size, min_dim;
	MPFMatrix mpfa;
	MPFVector mpfb, mpfx, mpfans;
	mpf_t reps, aeps, relerr[3];
#endif
	long int ret_f, ret_d, ret_mpf;
	long int *row_ch, *col_ch;
#ifdef USE_IMKL
	lapack_int *row_ch_imkl;
#endif
	long int i, j;

/* Double */
//	dim = 128;
	if(argc <= 1)
	{
#ifdef USE_GMP
		fprintf(stderr, "Usage: %s [dim] [prec_b]\n", argv[0]);
#else // USE_GMP
		fprintf(stderr, "Usage: %s [dim]\n", argv[0]);
#endif // USE_GMP
		return 0;
	}
	dim = atol(argv[1]);
	if(dim <= 0)
		return 0;

#ifdef USE_GMP
	prec = 128;
	if(argc >= 3)
	{
		prec = (unsigned long)atol(argv[2]);
		if(prec < 128)
			prec = 128;
	}
	goto mpfstart;

#endif // USE_GMP

	/* initialize */
	row_ch = (long int *)calloc(sizeof(long int), dim);
	col_ch = (long int *)calloc(sizeof(long int), dim);
	da = init_dmatrix(dim, dim);
	db = init_dvector(dim);
	dx = init_dvector(dim);
	dans = init_dvector(dim);

	/* get problem */
	get_dproblem(da, db, dans);

//	print_dmatrix(da);

	/* run DLUdecomp & SolveDLS */
	// ret_d = DLUdecomp(da);
	// ret_d = DLUdecompP(da, row_ch);
	// ret_d = DLUdecompC(da, row_ch, col_ch);
	stime = get_secv();
//	ret_d = DLUdecomp_strassen(da, 3);
	ret_d = DLUdecomp_strassen(da, 512);
	etime[0] = get_secv() - stime;

	printf("strassenLU: %f\n", etime[0]);
	//print_dmatrix(da);
	ret_d = SolveDLS(dx, da, db);
/*	printf("  i    row_ch[i]    col_ch[i]\n");
	for(i = 0; i < dim; i++)
		printf("%5ld %25.17e %25.17e\n", i, get_dvector_i(dx, i), get_dvector_i(dans, i));
*/

	get_dproblem(da, db, dans);
	stime = get_secv();
	ret_d = DLUdecomp(da);
	etime[1] = get_secv() - stime;
	printf("normalLU: %f\n", etime[1]);
	//print_dmatrix(da);
	
	ret_d = SolveDLS(dx, da, db);
	// ret_d = SolveDLSP(dx, da, db, row_ch);
	// ret_d = SolveDLSC(dx, da, db, row_ch, col_ch);

#ifdef USE_IMKL
	row_ch_imkl = (lapack_int *)calloc(sizeof(lapack_int), dim);

	get_dproblem(da, db, dans);
	stime = get_secv();
	ret_d = LAPACKE_dgetrf_work(LAPACK_ROW_MAJOR, (lapack_int)dim, (lapack_int)dim, da->element, (lapack_int)dim, row_ch_imkl);
	etime[2] = get_secv() - stime;
	printf("LAPACK LU: %f\n", etime[2]);
	//print_dmatrix(da);

	ret_d = LAPACKE_dgetrs_work(LAPACK_ROW_MAJOR, 'N', (lapack_int)dim, 1, da->element, (lapack_int)dim, row_ch_imkl, db->element, 1);
//	ret_d = SolveDLSP(dx, da, db, row_ch_imkl);

	/* print */
/*	printf("  i    row_ch[i]    col_ch[i]\n");
	for(i = 0; i < dim; i++)
		printf("%5ld %25.17e %25.17e\n", i, get_dvector_i(db, i), get_dvector_i(dans, i));
*/
	free(row_ch_imkl);
#endif

	/* end */
	free_dmatrix(da);
	free_dvector(db);
	free_dvector(dx);
	free_dvector(dans);
	free(row_ch);
	free(col_ch);

//	goto end;

#ifdef USE_GMP
/* MPF */
mpfstart:

	set_bnc_default_prec(prec);
	/* initialize */
	mpf_init(reps); mpf_init(aeps);
	mpf_init(relerr[0]);
	mpf_init(relerr[1]);
	mpf_init(relerr[2]);
	mpfa = init_mpfmatrix(dim, dim);
//	mpfa = init2_mpfmatrix(dim, dim, 256);
	mpfb = init_mpfvector(dim);
//	mpfb = init2_mpfvector(dim, 256);
	mpfx = init_mpfvector(dim);
//	mpfx = init2_mpfvector(dim, 256);
	mpfans = init_mpfvector(dim);

	/* get problem */
	get_mpfproblem(mpfa, mpfb, mpfans);

//	print_mpfmatrix(mpfa);

	/* run MPFLUdecomp & SolveMPFLS */
	stime = get_secv();
	ret_mpf = MPFLUdecomp(mpfa);
	etime[4] = get_secv() - stime;
	// ret_mpf = MPFLUdecompP(mpfa, row_ch);
	// ret_mpf = MPFLUdecompC(mpfa, row_ch, col_ch);
	ret_mpf = SolveMPFLS(mpfx, mpfa, mpfb);
	printf("normalLU(%ld, %ld): %f\n", dim, get_bnc_default_prec(), etime[4]);
	// ret_mpf = SolveMPFLSP(mpfx, mpfa, mpfb, row_ch);
	relerr_element_mpfvector(relerr[0], relerr[1], relerr[2], mpfx, mpfans, 0);
	printf("relerr(max, min, ||relerr||): ");
	mpf_out_str(stdout, 10, 3, relerr[0]); printf(", ");
	mpf_out_str(stdout, 10, 3, relerr[1]); printf(", ");
	mpf_out_str(stdout, 10, 3, relerr[2]); printf("\n");
	// ret_mpf = SolveMPFLSC(mpfx, mpfa, mpfb, row_ch, col_ch);

	min_dim = STRASSEN_MIN_DIM;
	etime[5] = 0.0;
//	for(block_size = min_dim; block_size <= min_dim * 4; block_size += min_dim)
	for(block_size = min_dim; block_size <= min_dim * 10; block_size += min_dim)
	{
		get_mpfproblem(mpfa, mpfb, mpfans);
		stime = get_secv();
		ret_mpf = MPFLUdecomp_strassen(mpfa, block_size);
	//	ret_mpf = MPFLUdecomp_strassen(mpfa, 64);
	//	ret_mpf = MPFLUdecomp_strassen(mpfa, 128);
	//	ret_mpf = MPFLUdecomp_strassen(mpfa, 16);
		etime[5] = get_secv() - stime;
		ret_mpf = SolveMPFLS(mpfx, mpfa, mpfb);
#ifdef USE_BLOCK 
		printf("blockLU(%ld, %ld, %ld): %f\n", dim, get_bnc_default_prec(), block_size, etime[5]);
#elif USE_STRASSEN
		printf("strassenLU(%ld, %ld, %ld): %f\n", dim, get_bnc_default_prec(), block_size, etime[5]);
#elif USE_WINOGRAD
		printf("winogradLU(%ld, %ld, %ld): %f\n", dim, get_bnc_default_prec(), block_size, etime[5]);
#endif
		relerr_element_mpfvector(relerr[0], relerr[1], relerr[2], mpfx, mpfans, 0);
		printf("relerr(max, min, ||relerr||): ");
		mpf_out_str(stdout, 10, 3, relerr[0]); printf(", ");
		mpf_out_str(stdout, 10, 3, relerr[1]); printf(", ");
		mpf_out_str(stdout, 10, 3, relerr[2]); printf("\n");
	}
	// relerr_element_mpfvector(relerr[0], relerr[1], relerr[2], mpfx, mpfans, 0);
	// printf("relerr(max, min, ||relerr||): ");
	//mpf_out_str(stdout, 10, 3, relerr[0]); printf(", ");
	//mpf_out_str(stdout, 10, 3, relerr[1]); printf(", ");
	//mpf_out_str(stdout, 10, 3, relerr[2]); printf("\n");

	/* print */
	/* for(i = 0; i < dim; i++)
	{
		printf("%5ld ", i);
		mpf_out_str(stdout, 10, 0, get_mpfvector_i(mpfx, i));
		printf(" ");
		mpf_out_str(stdout, 10, 0, get_mpfvector_i(mpfans, i));
		printf("\n");
	}
*/
	/* end */
	mpf_clear(reps); mpf_clear(aeps);
	mpf_clear(relerr[0]); mpf_clear(relerr[1]); mpf_clear(relerr[2]);
	free_mpfmatrix(mpfa);
	free_mpfvector(mpfb);
	free_mpfvector(mpfx);
	free_mpfvector(mpfans);
#endif

	/* print itimes */
end:
	return 0;

}

